package ant;

import java.util.*;
import ant.antinstructions.*;
import ObservableRegistry;
import BinaryParticlePosition;
import BinaryPSOFitnessEvaluator;

public class AntFSMTraverser implements BinaryPSOFitnessEvaluator {

    /**
     *  Bitset representing the FSM to traverse
     */
    private StateMachineBlueprint stateMachineBlueprint_;
    public StateMachineBlueprint stateMachineBlueprint() { return stateMachineBlueprint_; }
    public void stateMachineBlueprint( StateMachineBlueprint aStateMachineBlueprint ) { stateMachineBlueprint_ = aStateMachineBlueprint; }

    /**
     *  The instructions the ant can perform
     */
    private Vector instructions_;
    public Vector instructions() { return instructions_; }
    public void instructions( Vector aVector ) { instructions_ = aVector; }

    /**
     *  Position of the current state in the StateMachineBlueprint
     */
    private int currentState_;
    public int currentState() { return currentState_; }
    public void currentState( int anInt ) { currentState_ = anInt; }
    
    /**
     *  The FoodPlain associated with this ArtificialAntFitnessEvaluator
     */
    private FoodPlain foodPlain_;
    public FoodPlain foodPlain() { return foodPlain_; }
    public void foodPlain( FoodPlain aFoodPlain ) { foodPlain_ = aFoodPlain; }

    /**
     *  The body of the ant which acts as a slave to a
     *  chromosome associated with the evolved individual
     */
    private AntBody antBody_;
    public AntBody antBody() { return antBody_; }
    public void antBody( AntBody anAntBody ) { antBody_ = anAntBody; }


    /**
     *  The maximum number of instructions an ant
     *  can execute in a particular trial
     */
    private int executionLimit_;
    public int executionLimit() { return executionLimit_; }
    public void executionLimit( int anInt ) { executionLimit_ = anInt; }

    /**
     *  A request for this computer to halt at the first
     *  opportunity
     */
    private boolean haltIsRequested_;
    public boolean haltIsRequested() { return haltIsRequested_; }
    public void haltIsRequested( boolean aBoolean ) { haltIsRequested_ = aBoolean; }

    public AntFSMTraverser() {
        this.initialize();
    }
    
    private void initialize() {
        this.stateMachineBlueprint(null); // filled in by user
        this.initInstructions();
        this.currentState(0);
        this.foodPlain(new FoodPlain());
        this.antBody(new AntBody());
        this.executionLimit(550);
        this.haltIsRequested(false);

        ObservableRegistry.instance().register("foodPlain", this.foodPlain()); // allow observation
        
    }
    
    public void initInstructions() {
        this.instructions(new Vector());
        this.instructions().addElement(new MoveForward ());
        this.instructions().addElement(new FaceLeft    ());
        this.instructions().addElement(new FaceRight   ());
        this.instructions().addElement(new Nop         ());
        this.instructions().addElement(new Nop         ());
        this.instructions().addElement(new Nop         ());
        this.instructions().addElement(new Nop         ());
        this.instructions().addElement(new Nop         ());
    } 
    
    public synchronized void loadStateMachineBlueprint(StateMachineBlueprint theStateMachineBlueprint) {
        this.reset();
        this.haltIsRequested(false);
        this.stateMachineBlueprint(theStateMachineBlueprint);
        this.foodPlain().reset();
        this.antBody().reset();
    }
    
    public void abortExecution() {
        this.haltIsRequested(true);
    }
    
    public void reset() {
        this.currentState(0);
    }
    
    public boolean atExecutionLimit() {
        return (this.antBody().executionCount() >= this.executionLimit());
    }
    
    public Vector programListingFor(StateMachineBlueprint blueprint) {
        int q=0;
        Vector listing = new Vector();
        StringBuffer s = null;
        
        for (int i=0; i<8; i++) {
        //for (int i=0; i<4; i++) {
            s = new StringBuffer();
            int p = i*12; // point to state
            s.append("State " + i + " -");
            s.append( " FA ");
            //q = blueprint.fragmentAsInteger(p, p+2) % 4;
            q = blueprint.fragmentAsInteger(p, p+2);
            s.append(((AntInstruction) this.instructions().elementAt(q)).opCode());
            s.append(" Nxt ");
            s.append(blueprint.fragmentAsInteger(p+3, p+5));
            
            s.append( "   NFA ");
            //q = blueprint.fragmentAsInteger(p+6, p+8) % 4;
            q = blueprint.fragmentAsInteger(p+6, p+8);
            s.append(((AntInstruction) this.instructions().elementAt(q)).opCode());
            s.append(" Nxt ");
            s.append(blueprint.fragmentAsInteger(p+9, p+11));
            
            listing.addElement(s.toString());
        }
        return listing;
    }
                  
    public synchronized double execute() {
        double theFitness = 0;
        this.reset();
        
        while((!this.atExecutionLimit()) && (!this.haltIsRequested())) {
            Thread.yield(); // fixbg .. why 
            this.traverseNextLink();
        }

        return (double) this.antBody().foodEaten();
    }
    
    public void traverseNextLink() {
        
        AntInstruction theInstruction = this.nextInstruction();
        theInstruction.executeOn(this.foodPlain(), this.antBody());
        
        // let the world know what is happening
        this.foodPlain().publicize(this.antBody());
    }
    
    
      public AntInstruction nextInstruction() {
                
        // find current state in the blueprint
        int p = 12 * this.currentState();
        
        // find the link based on the input
        if (this.antBody().ifFoodAheadOn(this.foodPlain())) {
            // firstLink
            p += 0; // for clarity only
        }
        else {
            // second link
            p += 6;
        }
        
        // extract the info
        /*
        
        int outputSelector = intFromBitSetFragment(this.stateMachineBlueprint().bitSet(), p, p+2); // fixbg pretty this up ?
        int nextState = intFromBitSetFragment(this.stateMachineBlueprint().bitSet(), p+3, p+5);
        */
        
        int outputSelector = this.stateMachineBlueprint().fragmentAsInteger(p,   p+2);
        int nextState      = this.stateMachineBlueprint().fragmentAsInteger(p+3, p+5);
        
        // now execute the link output
        
        AntInstruction theNextInstruction = (AntInstruction) this.instructions()
                                      .elementAt(outputSelector);
                                    //  .elementAt(outputSelector % 4);
                                      
        
        // update the current state
        this.currentState(nextState);
        //this.currentState(nextState % 4);
                                      
        return theNextInstruction;
    }
    public int intFromBitSetFragment(BitSet b, int begin, int end) {
        // converts fragment of a BitSet into an int
        // begin and end are inclusive
        int positionValue = 1;
        int value = 0;
        for (int i=end; i>=begin; i--) {
            if (b.get(i)) {
                value += positionValue;
            }
            positionValue *= 2;
        }
        return value;
    }
    
    // BinaryPSOFitnessEvaluator interface methods
    
    public double evaluateFitnessFor(BinaryParticlePosition position) {
        
        this.loadStateMachineBlueprint((StateMachineBlueprint) position);
        double fitness = this.execute();
        return 89-fitness;  // zero is best fitness
    }
    
    public Vector descriptionOf(BinaryParticlePosition position) {
        return this.programListingFor((StateMachineBlueprint) position);
    }
        
}